
class Procs ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `DimensionRep': representation class for process array     //
//                          dimension.                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimensionRep {
public :
  int size() const {return extent ; }
  int crd() const {return coord ; }
  DimensionRep(const int _extent, const int _coord, const int _stride,
               const int _axis) :
    extent(_extent), coord(_coord), stride(_stride), mask(1 << _axis) {}

  void show() const {
    AdlibPrintf("DimensionRep :: show: size = %d crd = %d str_prc = %d "
                "dims = %d\n", extent, coord, stride, mask) ;
  }

private :
  int extent ;
  int stride ;
  int coord  ;

  int mask ;

friend class DimensionSet ;        // accesses `mask'
friend class Dimension ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Dimension': the process array dimension.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimensionSet ;

class Dimension {
public :
  // constructors and destructors

  Dimension() {rep = (DimensionRep*) 0 ; }

//wjp add 1998-3-15
//  Dimension(int w) {rep = new DimensionRep(1,0,1,w) ; }


  // methods

  int size() const {return rep ? rep->size() : 1 ; }
  int crd()  const {return rep ? rep->crd()  : 0 ; }

  //DimensionSet dims() const ;

  int str_prc() const {return rep->stride ; }

//  void where() const ;
//  void erewh() const ;

  // operators

  operator DimensionRep*() const {return rep ; }

  // debugging

  void show () const {
    if(rep)
      rep->show() ;
    else
      AdlibPrintf("Dimension :: show: Null dimension\n") ;
  }

private :
  DimensionRep* rep ;

  void create(const int _extent, const int _coord, const int _stride,
              const int _axis) {
    rep = new DimensionRep(_extent, _coord, _stride, _axis) ;
  }
  void destroy() {delete rep ; }

friend class Procs ;               // accesses `create', `destroy'.
friend class DimensionSet ;        // accesses `rep'
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `DimensionSet': A set of process array dimensions.         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimensionSet {
public :
  // constructors and destructors

  DimensionSet() : mask(0) {}

  // methods

  int member(Dimension d) const {return (mask & d.rep->mask) != 0 ; }

  int contains(DimensionSet ds) const {
    return (~mask & ds.mask) == 0 ;
  }

//  void where() const ;
//  void erewh() const ;

  // operators

  operator int() const {return mask ; }

  DimensionSet operator+(DimensionSet ds) const {
    return DimensionSet(mask | ds.mask) ;
  }

  DimensionSet operator-(DimensionSet ds) const {
    return DimensionSet(mask & ~ds.mask) ;
  }

  DimensionSet operator+(Dimension d) const {
    if(d.rep)
      return DimensionSet(mask | d.rep->mask) ;
    else
      return *this ;
  }

  DimensionSet operator-(Dimension d) const {
    if(d.rep)
      return DimensionSet(mask & ~(d.rep->mask)) ;
    else
      return *this ;
  }

  DimensionSet& operator-=(DimensionSet ds) {
    mask &= ~ds.mask ;
    return *this ;
  }

  DimensionSet& operator+=(DimensionSet ds) {
    mask |= ds.mask ;
    return *this ;
  }

  DimensionSet& operator-=(Dimension d) {
    if(d.rep) mask &= ~(d.rep->mask) ;
    return *this ;
  }

  DimensionSet& operator+=(Dimension d) {
    if(d.rep) mask |= d.rep->mask ;
    return *this ;
  }

//  DimensionSet& operator-=(long w) {
//     mask &= w;
//    return *this ;
//  }

private :
  DimensionSet(const long _mask) : mask(_mask) {}

  long mask ;

//friend class Dimension ;        // accesses `DimensionSet(const long _mask)'
} ;


//inline DimensionSet Dimension :: dims() const {
//  return rep ? DimensionSet(rep->mask) : DimensionSet() ;
//}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    struct `Coord'.                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Coord {
  //Coord() {}
  //Coord(Dimension _dim, const int _crd) : dim(_dim), crd(_crd) {}

  Dimension dim ;
  int crd ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Procs': the process array.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Group ;

class Procs {
public :

  // constructors and destructors

  Procs(const int _rank, int* _n, int* subIds = 0) ;
  ~Procs() ;

private :
  Procs() ;  // Only used in initialisation
  Procs(const Procs& p) ;  // forbid copying.

  // methods

public :
  int rnk() const {return rank ; }

  Dimension dim(const int i) const {
    if(i < 0 || i >= rank)
      AdlibError("Procs :: dim: axis out of range\n") ;
    return dimensions [i] ;
  }

  DimensionSet dims() const {return dimSet ; }

  int member() const {return isLocal ; }
  
  int size() const {return mySize ; }

  int id() const {return myId ; }

  int id_abs(const int id) const {return physIds [id] ; }

  int lead_abs() const {return physIds [0] ; }

//  int  on() const ;
//  void no() const ;
//
  // operators

  Group operator/(const Coord& i) const ;

private :
  int rank ;
  Dimension* dimensions ;
  int* physIds ;
  int isLocal ;
  int myId ;
  int mySize ;

  DimensionSet dimSet ;

friend void AdlibInit(int& argc, char** (&argv)) ;  // accesses `Procs()'
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Group'.                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Group {
public :
  // constructors

  Group() {}

  Group(const Procs& p) :
      procs(&p), dimSet(p.dims()), leadId(0), isLocal(p.member()) {}

  // methods

  const Procs* prc() const {return procs ; }

  DimensionSet dims() const {return dimSet ; }

  int member() const {return isLocal ; }

  int size() const ;

  int id() const ;

//wjp add
  void sub_id(int rank, int* g_index, int* sub_ids, int* l, int* u, int* s, int* inde, int* rk) const ;
  int check_sub(int rank, int* l, int* u) const ;

  int id_prc(const int id) const ;
  int id_abs(const int id) const {return procs->id_abs(id_prc(id)) ; }

  int lead_prc() const {return leadId ; }
  int lead_abs() const {return procs->id_abs(leadId) ; }

  void restrict(Dimension d, const int coord) ;
  void restrict(Dimension d) ;
  void restrict(DimensionSet ds) ;

//wjp add
//  void restrict_all() ;

//  int  on() const ;
//  void no() const ;

  // operators

  Group operator/(const Coord& i) const {
    Group result(*this) ;
    result.restrict(i.dim, i.crd) ;
    return result ;
  }

  inline Group& operator/=(const Coord& i) {
    restrict(i.dim, i.crd) ;
    return *this ;
  }

  // debugging

  void show() {
    AdlibPrintf("Group :: show: prc = %p dims = %x lead_prc = %d member = %d\n",
                procs, (int) dimSet, leadId, isLocal) ;
  }
protected :
  const Procs* procs ;

  DimensionSet dimSet ;
  int leadId ;
  int isLocal ;
} ;


inline Group Procs :: operator/(const Coord& i) const {
  Group result(*this) ;
  result /= i ;
  return result ;
}


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
////    `GroupStk': A lightweight ``Stack'' class.                       //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class GroupStk {
//public :
//  void init() ;
//
//  void push(Group* val) {
//    if(n == capacity) extend(2 * capacity) ;
//
//    vals [n] = *val ;
//    n++ ;
//  }
//
//  void pop(Group* val) {
//    n-- ;
//    *val = vals [n] ;
//  }
//
//private :
//  int n, capacity ;
//  Group* vals ;
//
//  void extend(const int newCapacity) ;
//} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Group `apg': the process group executing this control thread.    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

extern Group apg ;

//extern GroupStk apgStack ;

//inline int Procs :: on() const {
//  apgStack.push(&apg) ;
//  apg = *this ;
//  return member() ;
//}
//
//inline void Procs :: no() const {
//  apgStack.pop(&apg) ;
//}
//
//inline int Group :: on() const {
//  apgStack.push(&apg) ;
//  apg = *this ;
//  return member() ;
//}
//
//inline void Group :: no() const {
//  apgStack.pop(&apg) ;
//}

//inline void Dimension :: where() const {
//  apgStack.push(&apg) ;
//  apg.restrict(*this) ;
//}
//
//inline void Dimension :: erewh() const {
//  apgStack.pop(&apg) ;
//}
//
//inline void DimensionSet :: where() const {
//  apgStack.push(&apg) ;
//  apg.restrict(*this) ;
//}
//
//inline void DimensionSet :: erewh() const {
//  apgStack.pop(&apg) ;
//}

class ProcTree ;

//void apgSpanTree(ProcTree* tree) ;
void spanTree(ProcTree* tree, const Group& p) ;


// Used to restrict group of value-accessible arguments to APG...

inline void restrictToAPG(Group& p) {
  if(apg.prc() == p.prc()) {
    DimensionSet external = p.dims() - apg.dims() ;
    if(external)
      p.restrict(external) ;  // Restrict group to APG.
  }
}


// ---------------------------------------------------------------------
//
// Things to do...
//
//    Remove `Dimension :: axis' and add `Dimension :: dims' returning
//    a singleton `DimensionSet' (or empty set for a null dimension).

